# This is a sample Python script.

# Press Umschalt+F10 to execute it or replace it with your code.
# Press Double Shift to search everywhere for classes, files, tool windows, actions, and settings.
import torch
from typing import Dict

import multiprocessing
import os
import sys

from dataclasses import dataclass

import docopt
import toml
from loguru import logger
from tqdm import tqdm

import regioneval.utility
from paperboat.configuration.presets import default_ocr, default_baseline_layout, modern_ocr
from paperboat.model_loader import LocalModelLibrary
from paperboat.processing import PipelineBuilder
from segmentation.postprocessing.data_classes import PredictionResult
# import ray

# Do not remove these imports, these are necessary for the steps to be found
from paperboat.pipeline.preprocessing import *
from paperboat.pipeline.segmentation import *
from paperboat.pipeline.ocr import *
from paperboat.pipeline.output import *

@dataclass
class BaselinePredictorResult:
    prediction: PredictionResult
    scaled_image: SourceImage


class BaselinePredictionActor:
    def __init__(self, settings: 'PredictionSettings'):
        #if not model_path or (model_weights is not None):
        #    raise NotImplementedError("Loading model by weights is not implemented")
        self.settings = settings
        self.predictor = Predictor(settings)
        logger.info("Loaded baseline model successfully")

    def process(self, img: SourceImage) -> BaselinePredictorResult:
        prediction, scaled_image = self.predictor.predict_image(img)
        #, process_pool=multiprocessing.Pool()) # TODO: should this be parallelized???
        # maybe add this parallelization to the settings?
        return BaselinePredictorResult(prediction, scaled_image)

@dataclass
class LayoutProcessingResult:
    analyzed_content: 'AnalyzedContent'

class LayoutProcessingActor:
    def __init__(self, layout_settings: LayoutProcessingSettings):
        self.settings = layout_settings

    def process(self, bl_pred: BaselinePredictorResult) -> LayoutProcessingResult:
        # TODO: should we use innner parallelism here?
        analyzed_content = process_layout(bl_pred.prediction, bl_pred.scaled_image, multiprocessing.pool.ThreadPool(1), self.settings)
        return LayoutProcessingResult(analyzed_content)



def test_pipeline(img_f):
    # run segmentation on that image
    settings = PredictionSettings(["/home/norbert/share/BaselineModEval/mod/model_136.torch"], 1000000, None, None)
    nn_predictor = Predictor(settings)
    logger.info("Model loaded successfully!")

    with regioneval.utility.Stopwatch("Prediction"):
        prediction, scaled_image = nn_predictor.predict_image(SourceImage.load(img_f), process_pool=multiprocessing.Pool())

    scale_factor = prediction.prediction_scale_factor
    layout_settings = LayoutProcessingSettings()

    with regioneval.utility.Stopwatch("Layout"):
        analyzed_content = process_layout(prediction, scaled_image, multiprocessing.Pool(), layout_settings)

    # layout_debugging(args, analyzed_content, scaled_image, file)

    # convert this back to the original image space
    analyzed_content = analyzed_content.to_pagexml_space(scale_factor)

    # run calamari ocr


class UsageGenerator:
    def __init__(self):
        self.usage ="""paperboat.
Fully automated Page Segmentation and OCR.

Usage:
    paperboat [options] <image_files>...
    paperboat binarize [options] <image_files>...
    paperboat segment [options] <image_files>...
    paperbaot ocr [options] <image_files>...
    paperboat load-models [--clear]


Options:
    -o=DIR, --output_dir=DIR        Set output directory
    -f=FMT, --format=FMT            Set output format [default: xml]
    -l <dir>, --lookup_dir <dir>    Set directory to look for auxiliary files (e.g. PageXML Files)
    -c <config>, --config <config>  Set file containing configuration options
    -p <preset>, --preset <preset>  Load a given preset
    -t <threads>                    Run on a given number of threads
    --no_gpu                        Disable GPU for Segmentation and OCR
    -v                              Verbose mode ( do not suppress output )
    --Dcalamari.model <path>        Set path to the OCR model (.ckpt.json)
    --Dsegmentation.model <path>    Set path to the segmentation model (.torch)
    --Doutput.write_bin             Write binary image to output directory
    --Doutput.write_src             Write source image to output directory
"""
    def append_presets(self, presets: Dict[str,str]):
        self.usage += "\nAvailable presets:\n"
        for k, v in presets.items():
            vd = toml.loads(v)
            self.usage += f"\t- {k}: {vd.get('general',dict()).get('desc','')}\n"

    def append_formats(self, formats):
        self.usage += "\nAvailable output formats:\n"
        for k, v in formats.items():
            self.usage += f"\t- {k}: {v[0]}\n"

    def get_usage(self):
        return self.usage

def parse_optargs(args, config: dict ):
    def set_dict(d, path, vn, v):
        steps = path.split(".")
        for s in steps:
            if s in d:
                d = d[s]
            else:
                d[s] = dict()
                d = d[s]
        d[vn] = v

    if args["--Dcalamari.model"]:
        set_dict(config,"step.calamari", "model_path", args["--Dcalamari.model"][0])
    if args["--Dsegmentation.model"]:
        set_dict(config, "step.segmentation", "model_path", args["--Dsegmentation.model"][0])

    if args["--Doutput.write_bin"]:
        set_dict(config,"output", "write_bin", True)
    if args["--Doutput.write_src"]:
        set_dict(config,"output", "write_src", True)

def assume_output_dir(args):
    if not args["--output_dir"]:
        logger.error("Specify an output directory using -o / --output_dir")
        return False
    if len(args["--output_dir"]) > 1:
        logger.error("Can only specify one output directory")
        return False
    # if there is an output directory, make sure it's empty
    else:
        od = Path(args["--output_dir"][0])
        if od.exists():
            if not od.is_dir():
                logger.error("--output_dir must be a directory")
                return False
            if not sum(1 for _ in od.iterdir()) == 0:
                logger.error("--output_dir must be empty")
                return False
        else:
            logger.info(f"Creating output directory: {od}")
            os.mkdir(od)

    return True

def make_output_format_pagexml(args):
    return PageXMLOutputToDir(output_dir=args["--output_dir"][0] if args["--output_dir"] else None, write_bin=args["--Doutput.write_bin"])

def make_output_format_format1(args):
    return Format1OutputToDir(output_dir=args["--output_dir"][0])

def main(args: dict = None):
    ug = UsageGenerator()
    presets = {"baseline": default_baseline_layout, "default": default_ocr, "modern": modern_ocr}
    ug.append_presets(presets)
    output_formats = {
        "xml": ("PageXML", make_output_format_pagexml),
        "format1": ("Format1", make_output_format_format1)
    }
    ug.append_formats(output_formats)

    if not args:
        args = docopt.docopt(ug.get_usage(), version="0.0.1")

    if args["--preset"]:
        preset = presets[args["--preset"][0]]
    else:
        preset = default_ocr

    if args["--config"]:
        with open(args["--config"][0], "r") as f:
            preset = f.read()



    if args["--no_gpu"]:
        # disable GPU as quickly as possible
        os.environ["CUDA_VISIBLE_DEVICES"] = ""

    if args["load-models"]:
        lmod = LocalModelLibrary()
        if args["--clear"]:
            lmod.clear()

        else:
            lmod.load_default()
        sys.exit(0)
    elif args["binarize"] or args["segment"] or args["ocr"]:
        logger.error("Right now, only the full pipeline is implemented")
        sys.exit(1)
        if not args["-o"]:
            logger.error("Speify an output directory using -o DIR")
            sys.exit(1)
        pb = PipelineBuilder.from_configuration(toml.loads(default_ocr))
        pb.add(PageXMLOutputToDir(output_dir=args["-o"]))
        pipe = pb.get_pipeline()
        pipe.run(args["<image_files>"][0])
    # do not require an output dir to run, we can also run in place
    else:
        #if not assume_output_dir(args):
        #    sys.exit(1)
        config = toml.loads(preset)
        parse_optargs(args, config)
        del config["system"]["cuda_devices"]
        pb = PipelineBuilder.from_configuration(config)
        if args["--format"]:
            pb.add(output_formats[args["--format"][0]][1](args))
        else:
            pb.add(output_formats["xml"][1](args))

        pipe = pb.get_pipeline()
        for f in tqdm(args["<image_files>"]):
            try:
                if not args["-v"]:
                    pipe.run_suppressed(f)
                else:
                    pipe.run(f)
            except Exception as e:
                logger.critical(f"An error occured processing file: {f}")
                logger.critical(f"Error: {e.__str__()}")
                logger.critical(f"Aborting")
                raise e

        pipe.metrics.print_metrics()

    # run the default pipeline

if __name__ == '__main__':
    main()


