from tfaip import PipelineMode

from calamari_ocr.ocr.dataset.datareader.pagexml.reader import PageXMLDatasetLoader, PageXMLReader
from calamari_ocr.ocr.predict.params import PredictorParams
from pathlib import Path

import imageio
import tempfile

import typing

from paperboat.model_loader import ModelRetriever
from paperboat.processing import Step, CarryOnData
from regioneval.libpagexml2 import PageXMLParser, PageXMLParserMode, PageXMLBuilder, TextEquiv
from segmentation.preprocessing.source_image import SourceImage
from pathlib import PosixPath

class CalamariQueue:
    def __init__(self):
        self.queue = []

    def __iter__(self):
        return iter(self.queue)

    def append(self, t: object):
        self.queue.append(t)

class CalamariOCR(Step):
    def __init__(self, model_path: typing.List[str]):
        from tfaip import PipelineMode
        from tfaip.data.databaseparams import DataPipelineParams

        from calamari_ocr.ocr.voting import VoterParams

        from calamari_ocr.ocr.predict.predictor import MultiPredictor
        voter_params = VoterParams()
        voter_params.type = VoterParams.type.ConfidenceVoterDefaultCTC

        model_path = [(cp if cp.endswith(".json") else cp + ".json") for cp in model_path]
        model_path = [cp[:-5] for cp in model_path]

        self.nn_predictor = MultiPredictor.from_paths(
            checkpoints=model_path,
            voter_params=voter_params,
            predictor_params=PredictorParams(silent=True,
                                             progress_bar=True,
                                             pipeline=DataPipelineParams(batch_size=1,
                                                                         mode=PipelineMode("prediction"))
                                             )
        )
        #self.raw_predictor = self.nn_predictor.raw()
        #self.raw_pred_fn = self.raw_predictor.__enter__()  # TODO: exit?!?




    @classmethod
    def get_id(cls) -> str:
        return "calamari"

    def run(self, data: CarryOnData) -> CarryOnData:
        from calamari_ocr.ocr.dataset.datareader.pagexml.reader import PageXML
        source_image: SourceImage = data.get("source_image")
        with tempfile.TemporaryDirectory() as tmpdirname:
            img_file_name = Path(tmpdirname) / "img.png"
            imageio.imwrite(img_file_name, source_image.binarized(), compression=0)
            xml_file_name = Path(tmpdirname) / "img.xml"
            # write the xml file

            with open(xml_file_name,"w") as xml_file:
                xml_parse = PageXMLParser(data.get("layout.pxml"), filename="img.png",
                                          strictness=PageXMLParserMode.EASY).get_page()
                new_pxml = PageXMLBuilder.build_pagexml(xml_parse)
                xml_file.write(new_pxml)

            data_params = PageXML(images=[str(img_file_name)], xml_files=[str(xml_file_name)])
            predictor = self.nn_predictor.predict(data_params)
            for i in predictor:
                # get the id
                tl_id = i.meta["id"].split("/")[-1]
                xml_parse.textline_by_id(tl_id).text_equivs[TextEquiv.TEXT_ID_PREDICTION] = TextEquiv(TextEquiv.TEXT_ID_PREDICTION, i.outputs[1].sentence)


        data.set("calamari.page_out", xml_parse)
        return data

    def input_format(self) -> typing.Any:
        return CarryOnData

    def output_format(self) -> typing.Any:
        return CarryOnData

    @classmethod
    def from_config(cls, config: dict) -> 'CalamariOCR':
        # we have to load the model here
        # and get the prediction settings
        if config.get("model_path") is not None:
            model_path = config.get("model_path")
        elif config.get("repo_model") is not None:
            model_path = ModelRetriever({}).get_lib().get(config.get("repo_model"))
        else:
            model_path = ModelRetriever({}).get_lib().get_ocr_model()

        return CalamariOCR([str(model_path)])
