import imageio
import json
from dataclasses import dataclass
from pathlib import Path

import typing

from paperboat.pipeline.formatone import pagexml_to_format1
from paperboat.processing import CarryOnData, Step
from regioneval.libpagexml2 import PageXMLPage, PageXMLBuilder, PageXMLParser, PageXMLParserMode
from segmentation.preprocessing.source_image import SourceImage


@dataclass
class PageXMLOutput:
    image: SourceImage
    page: PageXMLPage
    pagexml: str



class PageXMLOutputToDir(Step):
    def __init__(self, output_dir=None, write_bin=False):
        self.output_dir = Path(output_dir) if output_dir else None
        self.write_bin = write_bin

    def input_format(self) -> typing.Any:
        return CarryOnData # we accept filenames as string

    def output_format(self) -> typing.Any:
        return PageXMLOutput

    @classmethod
    def from_config(cls, config: dict) -> 'PageXMLOutputToDir':
        if not "output_dir" in config:
            raise KeyError("Must have configuration option output_dir")
        return PageXMLOutputToDir(config.get("output_dir", ""))

    def run(self, data: CarryOnData) -> PageXMLOutput:
        # load the image
        if data.has("calamari.page_out"):
            page: PageXMLPage = data.get("calamari.page_out")

        else:
            page = PageXMLParser(data.get("layout.pxml"), strictness=PageXMLParserMode.EASY).get_page()
        pagexml = PageXMLBuilder.build_pagexml(page)
        # write that to the directory
        if self.output_dir:
            output_path = (self.output_dir / page.image_filename).with_suffix(".xml")
        else:
            output_path = (Path(data.get("src.filename")).parent / page.image_filename).with_suffix(".xml")

        with open(output_path, "w") as f:
            f.write(pagexml)

        if self.write_bin:
            src_image: SourceImage = data.get("source_image")
            imageio.imwrite((self.output_dir / page.image_filename).with_suffix(".png"),src_image.binarized(),compression=1)
        return PageXMLOutput(image=None, page=page, pagexml=pagexml)


    @classmethod
    def get_id(cls) -> str:
        return "pagexmloutputtodir"


class Format1OutputToDir(Step):
    def __init__(self, output_dir):
        self.output_dir = Path(output_dir)

    def input_format(self) -> typing.Any:
        return CarryOnData # we accept filenames as string

    def output_format(self) -> typing.Any:
        return PageXMLOutput

    @classmethod
    def from_config(cls, config: dict) -> 'Format1OutputToDir':
        if not "output_dir" in config:
            raise KeyError("Must have configuration option output_dir")
        return Format1OutputToDir(config["output_dir"])

    def run(self, data: CarryOnData) -> PageXMLOutput:
        # load the image
        if data.has("calamari.page_out"):
            page: PageXMLPage = data.get("calamari.page_out")

        else:
            page = PageXMLParser(data.get("layout.pxml"), strictness=PageXMLParserMode.EASY).get_page()

        f1page = pagexml_to_format1(page)
        # write that to the directory
        with open((self.output_dir / page.image_filename).with_suffix(".json"), "w") as f:
            json.dump(f1page.to_dict(), f, indent=2)
        return None


    @classmethod
    def get_id(cls) -> str:
        return "format1outputtodir"

