import enum

import contextlib
import io
import loguru
from contextlib import suppress

import os

import imageio
import sys
import toml

from abc import ABC, abstractmethod, abstractclassmethod

import typing
from loguru import logger
from typing import List
from dataclasses import dataclass

from paperboat.pipeline.pipeline_metrics import PipelineMetrics


class InvalidStepConfiguration(Exception):
    pass

class Step(ABC):
    @classmethod
    @abstractmethod
    def get_id(cls) -> str:
        pass

    @abstractmethod
    def run(self, data: dict) -> dict:
        raise NotImplementedError("Abstract method called")

    @abstractmethod
    def input_format(self) -> typing.Any:
        raise NotImplementedError("Abstract method called")

    @abstractmethod
    def output_format(self) -> typing.Any:
        raise NotImplementedError("Abstract method called")

    @classmethod
    @abstractmethod
    def from_config(cls, config: dict) -> 'Step':
        raise NotImplementedError("Abstract method called")

    @classmethod
    def get_implementation(cls, id: str) -> typing.Type['Step']:
        try:
            return next(c for c in cls.__subclasses__() if c.get_id() == id)
        except StopIteration:
            raise KeyError(id)


class ProcessingDataset:
    pass


class CarryOnData:
    def __init__(self):
        self.data = dict()

    def get(self, id: str):
        return self.data[id]

    def set(self, id: str, data):
        self.data[id] = data
        return self

    def has(self, id: str) -> bool:
        return id in self.data

@dataclass
class SystemSettings:
    beacon: bool = False
    cuda_devices: str = ""

    def apply(self):
        if self.cuda_devices != "auto":
            logger.warning(f"Setting Cuda Visible Devices to: {self.cuda_devices}")
            os.environ["CUDA_VISIBLE_DEVICES"] = self.cuda_devices
        self.beacon = True

    def warn_if_not_applied(self):
        if not self.beacon:
            logger.warning(f"Did not apply system Settings!")

@dataclass
class ProcessingPipeline:
    steps: List[Step]
    metrics: PipelineMetrics = PipelineMetrics()

    def run(self, data):
        v = data
        for step in self.steps:
            with self.metrics.get_counter(step):
                v = step.run(v)
        return v

    def run_suppressed(self, data):
        logger.level("ERROR")
        loguru.logger.level("ERROR")
        """
        with open(os.devnull, 'w') as nd:
            with contextlib.redirect_stdout(nd):
                with contextlib.redirect_stderr(nd):
        """
        res = self.run(data)
        loguru.logger.level("INFO")
        logger.level("INFO")
        return res



class PipelineBuilder:
    def __init__(self, config, desc="", auto_apply_settings=True):
        self.steps : List[Step] = []
        self.desc = desc
        self.config = config
        self.system_settings = SystemSettings()
        with suppress(KeyError):
            if not self.config["system"]["auto_cuda"]:
                self.system_settings.cuda_devices = self.config["system"]["cuda_devices"]
            else:
                self.system_settings.cuda_devices = "auto"

        if auto_apply_settings:
            self.system_settings.apply()


    def add(self, step: Step):
        if len(self.steps) > 0:
            # make sure that this step is compatible to the step before
            if not (step.input_format() == self.steps[-1].output_format()):
                logger.error(f"Invalid transport format between step {self.steps[-1].__class__.get_id()} and {step.__class__.get_id()}")
                logger.error(f"Step {step.__class__.get_id()} requires {step.input_format()} but got {self.steps[-1].output_format()}")
        self.steps.append(step)

    def add_many(self, steps: List[Step]):
        for x in steps:
            self.add(x)

    def get_pipeline(self):
        self.system_settings.warn_if_not_applied()
        return ProcessingPipeline(self.steps)

    @staticmethod
    def from_configuration(config: dict) -> 'PipelineBuilder':
        steps = config.get("general",{}).get("steps",[])
        pb = PipelineBuilder(config=config, desc=config.get("general",{}).get("desc",""))
        for step_id in steps:
            # make that step
            try:
                StepCls = Step.get_implementation(step_id)
            except KeyError as e:
                logger.error(f"Cannot find implementation for step {step_id}")
                logger.error(f"Check your pipeline configuration")
                sys.exit(1)
            # load step from config
            try:
                step = StepCls.from_config(config["step"][step_id])
            except KeyError as e:
                logger.error(f"configuration file has no section for step {step_id}")
                sys.exit(1)
            except InvalidStepConfiguration as e:
                logger.error(f"Invalid configuration for step {step_id}")
                logger.error(e)
                sys.exit(1)
            pb.add(step)
        return pb
